#include <stdio.h>
#include <stdlib.h>

#pragma pack(push, 1)
typedef struct {
	unsigned long	dwMagic;
	unsigned long	dwSize;
	unsigned long	dwFlags;
	unsigned long	dwHeight;
	unsigned long	dwWidth;
	unsigned long	dwPitchOrLinearSize;
	unsigned long	dwDepth;
	unsigned long	dwMipMapCount;
	unsigned long	dwReserved1[11];
	unsigned long	dwPfSize;
	unsigned long	dwPfFlags;
	unsigned long	dwFourCC;
	unsigned long	dwRGBBitCount;
	unsigned long	dwRBitMask;
	unsigned long	dwGBitMask;
	unsigned long	dwBBitMask;
	unsigned long	dwRGBAlphaBitMask;
	unsigned long	dwCaps;
	unsigned long	dwCaps2;
	unsigned long	dwReservedCaps[2];
	unsigned long	dwReserved2;
} DDSHEAD;
#pragma pack(pop)

#define DDSD_HEIGHT 	 0x00000002
#define DDSD_WIDTH 	 0x00000004
#define DDSD_PIXELFORMAT 	 0x00001000

#define DDPF_ALPHAPIXELS 	 0x00000001
#define DDPF_ALPHA 	 0x00000002
#define DDPF_RGB 	 0x00000040

void thtx_to_dds(char *tx, int len, char *outname)
{
	FILE *out;
	DDSHEAD hed;
	int i;
	char c;
	memset(&hed, 0, sizeof(DDSHEAD));

	if(memcmp(tx, "THTX", 4)) return;
	hed.dwMagic = ' SDD';
	hed.dwSize = 124;
	hed.dwFlags =  DDSD_HEIGHT | DDSD_WIDTH | DDSD_PIXELFORMAT;
	hed.dwHeight = *(short*)&tx[10];
	hed.dwWidth = *(short*)&tx[8];
	hed.dwPfSize = 32;
	hed.dwPfFlags =  DDPF_RGB | DDPF_ALPHAPIXELS;
	switch(*(short*)&tx[6]) {
	case 1:
		hed.dwRGBBitCount = 32;
		hed.dwRBitMask = 0x00ff0000;
		hed.dwGBitMask = 0x0000ff00;
		hed.dwBBitMask = 0x000000ff;
		hed.dwRGBAlphaBitMask = 0xff000000;
		break;
	case 5: 
		hed.dwRGBBitCount = 16;
		hed.dwRBitMask = 0x00000f00;
		hed.dwGBitMask = 0x000000f0;
		hed.dwBBitMask = 0x0000000f;
		hed.dwRGBAlphaBitMask = 0x0000f000;
		break;
	default: return;
	}

	out = fopen(outname, "wb");
	if(!out) return;
	fwrite(&hed, sizeof(DDSHEAD), 1, out);
	fwrite(tx + 0x10, 1, len - 0x10, out);
	fclose(out);
}

int main(int argc, char *argv[])
{
	FILE *fp;
	char head[0x40];
	char name[0x20];
	char ddsname[64];
	char *entbuf;
	long filelen;
	unsigned long ver;
	unsigned long entlen;
	unsigned long entofs;
	unsigned long nextofs;
	unsigned long nameofs;
	unsigned long baseofs;
	int i;

	if (argc != 2) {
		printf(
			"usage: anm2dds FILE\n"
			"Convert Touhou Animation FILE to DDS\n");
		return 0;
	}
	
	fp = fopen(argv[1], "rb");
	if(!fp) {
		perror("fopen");
		return 0;
	}
	fseek(fp, 0, SEEK_END);
	filelen = ftell(fp);
	rewind(fp);

	baseofs = 0x0;
	i = 0;
	do {
		/* wb_ǂ */
		if(fseek(fp, baseofs, SEEK_SET)) {
			perror("fseek");
			fclose(fp);
			return 0;
		}
		if(fread(&head, 1, 0x40, fp) != 0x40) {
			perror("fread");
			fclose(fp);
			return 0;
		}
		/* o[W`FbN */
		ver = *(int*)&head[0x28];
		if(ver != 2 && ver != 3 && ver != 4) {
			fprintf(stderr, "version not supported");
			fclose(fp);
			return 0;
		}
		/* Of[^ǂ */
		nameofs = *(int*)&head[0x1c];
		if(fseek(fp, baseofs + nameofs, SEEK_SET)) {
			perror("fseek");
			fclose(fp);
			return 0;
		}
		if(fread(name, 1, 0x20, fp) != 0x20) {
			perror("fread");
			fclose(fp);
			return 0;
		}
		entofs = *(int*)&head[0x30];
		if(baseofs + entofs >= filelen) {
			fprintf(stderr, "malformed entry header");
			fclose(fp);
			return 0;
		}
		nextofs = *(int*)&head[0x38];
		if(baseofs + nextofs + 0x3B >= filelen) {
			fprintf(stderr, "malformed entry header");
			fclose(fp);
			return 0;
		}
		if(nextofs != 0)
			entlen = nextofs - entofs;
		else
			entlen = filelen - entofs - baseofs;
		if(fseek(fp, baseofs + entofs, SEEK_SET)) {
			perror("fseek");
			fclose(fp);
			return 0;
		}

		entbuf = malloc(entlen);
		if(!entbuf) {
			perror("malloc");
			fclose(fp);
			return 0;
		}
		if(fread(entbuf, 1, entlen, fp) != entlen) {
			perror("fread");
			free(entbuf);
			fclose(fp);
			return 0;
		}
		{
			char *p = strrchr(name, '/');
			if(p) {
				strcpy(ddsname, p + 1);
			} else {
				strcpy(ddsname, name);
			}
			p = strrchr(ddsname, '.');
			if(p) {
				strcpy(p + 1, "dds");
			} else {
				strcat(ddsname, ".dds");
			}
		}
		printf("converting %s to %s...\n", name, ddsname);
		thtx_to_dds(entbuf, entlen, ddsname);
		free(entbuf);

		baseofs += nextofs;
	} while (nextofs);

	fclose(fp);
}/*

@@@@@mi@@_,A,,........_@
@@@@@܁@@>,T'--M.A
@@@@@@@ @i@LL؁v
@@@ ,@@@@@|,iq  -Ɂv@@@@.@@.
@@@@@@@@@(=o=p, ; .'L@M.  ;@M
@ @@@@@@@@/ . 'i __,.'.;L," :L,L'@. ށ@.M@.
@.,,.,.,,,.,.,,,.,.,,,.,.,,,.,.,M'!,i, i.j************************/
