/*
 + DISCLAIMER
  USE AT YOUR OWN RISK.

 + TECHNICAL DESCRIPTION
  ܂Ɍ|\[Xe[uĂ̂͊ȗ̂߁H
  ǂݍ݂An[hR[fBO̕ςƎv̂łcc
  [] = crypted, {} = compressed
  <Header>
  Algo: TH_CRYPTER
  TH_CRYPTER param: key=1Fh step=27h block=0Ch limit=400h
  +00 <magic 'PBGZ'>.4B
  [
    +04 <file count>.L + 123456
    +08 <offset to list>.L + 345678
    +0c <original size of list>.L + 567891
  ]

  <Data>
  Algo: TH_LZ -> TH_CRYPTER
  TH_CRYPTER param: crypt_param[<type>]
  {
    +00 <magic 'edz'>.3B
    +03 <type>.B
    [ +04 <data>.?B ]
  }

  <List>
  Algo: TH_LZ -> TH_CRYPTER
  TH_CRYPTER param: key=3Eh step=9Bh block=80h limit=400h
  {
    [ (
      +00 <null terminated filename>.?B
      +XX <offset to data>.L
      +YY <original size of data>.L
      +ZZ <unknown data>.L
      ) * <file count>
    ]
  }
*/

#include <stdio.h>
#include <stdlib.h>

void th_crypter(unsigned char* in, unsigned char *out, int size, unsigned char key, unsigned char step, int block, int limit)
{
	unsigned char *i_itr = in;
	unsigned char *o_itr = out;
	unsigned char *o_base;
	int addup;
	int i;

	addup = size % block;
	if(addup >= block / 4) addup = 0;
	addup += size % 2;
	size -= addup;

	while(size > 0 && limit > 0) {
		if(size < block) block = size;
		/* first phase */
		o_base = o_itr;
		o_itr = o_itr + block - 1;
		for(i = (block + 1) / 2; i > 0; --i, ++i_itr) {
			/* decode char */
			*o_itr = *i_itr ^ key;
			o_itr -= 2;
			/* update a key */
			key += step;
		}
		/* second phase */
		o_itr = o_base + block - 2;
		for(i = block / 2; i > 0; --i, ++i_itr) {
			/* decode char */
			*o_itr = *i_itr ^ key;
			o_itr -= 2;
			/* update a key */
			key += step;
		}
		/* seek pointers */
		size -= block;
		o_itr = o_base + block;
		limit -= block;
	}
	/* transfer remaining data */
	size += addup;
	if(size > 0) memcpy(o_itr, i_itr, size);
}

struct bitstrm {
	unsigned char b;
	unsigned char c;
	unsigned char *cur;
	unsigned char *base;
	int size;
};

void bitstrm_init(struct bitstrm *strm, unsigned char *buf, int size)
{
	strm->b = 0x80;
	strm->c = 0;
	strm->cur = strm->base = buf;
	strm->size = size;
}

int bitstrm_read(struct bitstrm *strm, int bits)
{
	int ret = 0;
	int i;

	for(i = 1 << (bits - 1); i > 0; i >>= 1) {
		if(strm->b == 0x80) {
			if(strm->cur - strm->base < strm->size)
				strm->c = *strm->cur++;
			else /* idling */
				strm->c = 0;
		}
		if(strm->b & strm->c) ret |= i;
		strm->b >>= 1;
		if (!strm->b) strm->b = 0x80;
	}
	return ret;
}

/* TH08 LZ (18bit LZSS, Dictionary = 8KB)
 * 1.b <char>.B
 * 0.b <offset>.13b <length>.4b
 * 0.b 0.13b (EOF) */
#define DICT_SIZE 0x2000
void th_unlz(unsigned char *in, int insize, unsigned char *out, int outsize)
{
	unsigned char g_dict[DICT_SIZE];
	struct bitstrm bs;
	int dicofs = 1;
	int patofs;
	int patlen;
	int c, i;

	memset(g_dict, 0, DICT_SIZE);
	bitstrm_init(&bs, in, insize);

	while(outsize > 0) {
		/* flag (1 bit) */
		int flag = bitstrm_read(&bs, 1);
		if(flag) {	/* plain */
			/* plain char (8 bits) */
			c = bitstrm_read(&bs, 8);
			*out++ = c;
			g_dict[dicofs] = c;
			dicofs = (dicofs + 1) % DICT_SIZE;
			--outsize;
		} else {	/* compressed */
			/* offset to the pattern (13 bits) */
			patofs = bitstrm_read(&bs, 13);
			if(!patofs) break;
			/* length of the pattern - 2 (4 bits) */
			patlen = bitstrm_read(&bs, 4) + 2;
			/* decompress data */
			for(i = 0; i <= patlen && outsize > 0; ++i, --outsize) {
				c = g_dict[(patofs + i) % DICT_SIZE];
				*out++ = c;
				g_dict[dicofs] = c;
				dicofs = (dicofs + 1) % DICT_SIZE;
			}
		}
	}
}

struct th_file_entry {
	char *name;
	int offset;
	int complen;
	int origlen;
};

/* presets of crypter parameters */
typedef struct {
	unsigned char type;
	unsigned char key;
	unsigned char step;
	int block;
	int limit;
} CRYPT_PARAM;
/* for Touhou Eiyashou */
#define CRYPRM_MAX 8
CRYPT_PARAM g_cryprm_08[] = {
	{ 0x4d, 0x1b, 0x37, 0x40, 0x2000 },
	{ 0x54, 0x51, 0xe9, 0x40, 0x3000 },
	{ 0x41, 0xc1, 0x51, 0x1400, 0x2000 },
	{ 0x4a, 0x03, 0x19, 0x1400, 0x7800 },
	{ 0x45, 0xab, 0xcd, 0x200, 0x1000 },
	{ 0x57, 0x12, 0x34, 0x400, 0x2800 },
	{ 0x2d, 0x35, 0x97, 0x80, 0x2800 },
	{ 0x2a, 0x99, 0x37, 0x400, 0x1000 }
};
/* for Touhou Kaeizuka */
CRYPT_PARAM g_cryprm_09[] = {
	{ 0x4d, 0x1b, 0x37, 0x40,  0x2800 },
	{ 0x54, 0x51, 0xe9, 0x40,  0x3000 },
	{ 0x41, 0xc1, 0x51, 0x400, 0x400 },
	{ 0x4a, 0x03, 0x19, 0x400, 0x400 },
	{ 0x45, 0xab, 0xcd, 0x200, 0x1000 },
	{ 0x57, 0x12, 0x34, 0x400, 0x400 },
	{ 0x2d, 0x35, 0x97, 0x80,  0x2800 },
	{ 0x2a, 0x99, 0x37, 0x400, 0x1000 }
};

int th_extract_file(FILE *dat, struct th_file_entry *entry, int cryptver)
{
	CRYPT_PARAM *cryprm;
	FILE *fp;
	char *inbuf, *outbuf;
	int i;

	/* set cryprm */
	cryprm = (cryptver == 0) ? g_cryprm_08 : g_cryprm_09;

	/* allocate buffers for extracting data */
	inbuf = malloc(entry->complen);
	if(!inbuf) return 0; 
	outbuf = malloc(entry->origlen);
	if(!outbuf) {
		free(inbuf);
		return 0;
	}

	/* read and decompress data */
	fseek(dat, entry->offset, SEEK_SET);
	fread(inbuf, 1, entry->complen, dat);

	th_unlz(inbuf, entry->complen, outbuf, entry->origlen);
	free(inbuf);
	inbuf = outbuf;
	
	/* check magic */
	if(inbuf[0] != 'e' || inbuf[1] != 'd' || inbuf[2] != 'z') {
		free(inbuf);
		return 0;
	}

	/* check type */
	for(i = 0; i < CRYPRM_MAX && cryprm[i].type != inbuf[3]; ++i);
	if(i == CRYPRM_MAX) {
		free(inbuf);
		return 0;
	}

	/* decrypt data */
	outbuf = malloc(entry->origlen - 4);
	if(!outbuf) {
		free(inbuf);
		return 0;
	}
	th_crypter(inbuf + 4, outbuf, entry->origlen - 4,
		cryprm[i].key, cryprm[i].step, cryprm[i].block, cryprm[i].limit);
	free(inbuf);

	/* write data to the file */
	fp = fopen(entry->name, "wb");
	if (!fp) {
		free(outbuf);
		return 0;
	}
	fwrite(outbuf, 1, entry->origlen - 4, fp);
	free(outbuf);

	fclose(fp);

	return 1;
}

int main(int argc, char *argv[])
{
	FILE *fp;
	struct th_file_entry *filents;
	char *inbuf, *outbuf, *it, *list;
	int filelen;
	int complen;
	int magic;
	int files;
	int listpos;
	int listlen;
	int i;

	if(argc < 3) {
		printf(
		"usage: pbgzmlt <arch> <cmd> [<param> ...]\n"
		"Uncompress Touhou Eiyashou/Kaeizuka PBGZ arch\n\n"
		"cmd:\n"
		"  l\tlist files\n"
		"  eN\textract <param>\t(N=8:Eiyashou; N=9:Kaeizuka)\n"
		"  aN\textract all\t(N=8:Eiyashou; N=9:Kaeizuka)\n");
		exit(EXIT_SUCCESS);
	}

	fp = fopen(argv[1], "rb");
	if(fp == NULL) return 0;
	
	/* get the length of the file */
	fseek(fp, 0, SEEK_END);
	filelen = ftell(fp);
	rewind(fp);

	/* check magic */
	fread(&magic, 4, 1, fp);
	if(magic != 'ZGBP') {
		fprintf(stderr, "invalid file");
		fclose(fp);
		return 0;
	}
	
	/* allocate buffers for the header */
	inbuf = malloc(12);
	if(!inbuf) {
		perror("malloc");
		fclose(fp);
		return 0;
	}
	outbuf = malloc(12);
	if(!outbuf) {
		perror("malloc");
		free(inbuf);
		fclose(fp);
		return 0;
	}
	/* read and decrypt the header */
	if(fread(inbuf, 1, 12, fp) != 12) {
		fprintf(stderr, "invalid header");
		free(inbuf);
		free(outbuf);
		fclose(fp);
		return 0;
	}
	th_crypter(inbuf, outbuf, 12, 27, 55, 12, 1024);
	free(inbuf);

	files   = ((int*)outbuf)[0] - 123456;
	listpos = ((int*)outbuf)[1] - 345678;
	listlen = ((int*)outbuf)[2] - 567891;
	free(outbuf);

	/* verify the header element(s) */
	if(listpos > filelen) {
		fprintf(stderr, "malformed header");
		fclose(fp);
		return 0;
	}
	fseek(fp, listpos, SEEK_SET);

	/* allocate buffers for the list of files */
	complen = filelen - listpos;
	inbuf = malloc(complen);
	if(!inbuf) {
		perror("malloc");
		fclose(fp);
		return 0;
	}
	outbuf = malloc(complen);
	if(!outbuf) {
		perror("malloc");
		free(inbuf);
		fclose(fp);
		return 0;
	}
	/* read and decrypt the list */
	if(fread(inbuf, 1, complen, fp) != complen) {
		free(inbuf);
		free(outbuf);
		fclose(fp);
		return 0;
	}
	th_crypter(inbuf, outbuf, complen, 62, 155, 128, 1024);
	free(inbuf);

	/* decompress the list */
	inbuf = outbuf;
	list = malloc(listlen);
	if(!list) {
		perror("malloc");
		free(inbuf);
		fclose(fp);
		return 0;
	}
	th_unlz(inbuf, complen, list, listlen);
	free(inbuf);

	/* process the list */
	filents = malloc(files * sizeof(struct th_file_entry));
	if(!filents) {
		perror("malloc");
		free(list);
		free(filents);
		fclose(fp);
		return 0;
	}
	it = list;
	for(i = 0; it - list < listlen; ++i) {
		char *name, *tmp;
		/* check a name */
		for(tmp = it; tmp - list < listlen && *tmp; ++tmp);
		if(*tmp) {
			fprintf(stderr, "list may be corrupted (1)");
			free(list);
			free(filents);
			fclose(fp);
			return 0;
		}
		if (tmp - it == 0) break;

		if(++tmp + 4 * 3 - list >= listlen) {
			fprintf(stderr, "list may be corrupted (2)");
			free(list);
			free(filents);
			fclose(fp);
			return 0;
		}
		/* read and check an entry */
		filents[i].name = strdup(it);
		if(!filents[i].name) {
			perror("strdup");
			free(list);
			free(filents);
			fclose(fp);
			return 0;
		}
		it = tmp;
		filents[i].offset  = ((int*)it)[0];
		filents[i].origlen = ((int*)it)[1];
		it += 12;
		if (filents[i].offset >= filelen) {
			fprintf(stderr, "malformed file-entry");
			free(list);
			free(filents);
			return 0;
		}
	}
	free(list);

	/* calculate the lengths of compressed files */
	files = i;
	for(i = 0; i < files - 1; ++i) {
		filents[i].complen = filents[i + 1].offset - filents[i].offset;
		if(filents[i].origlen <= 0) {
			free(list);
			free(filents);
			fclose(fp);
			return 0;
		}
	}
	filents[i].complen = listpos - filents[i].offset;
	if(filents[i].complen <= 0) {
		free(list);
		free(filents);
		fclose(fp);
		return 0;
	}

	/* extract files */
	switch(argv[2][0]) {
	case 'l':
		printf("%-16s%16s%8s\n", "Filename", "Filesize", "Ratio");
		printf("----------------------------------------\n");
		for(i = 0; i < files; ++i) {
			printf("%-16s%#16d%#8d\n", filents[i].name, filents[i].origlen, filents[i].complen * 100 / filents[i].origlen);
			free(filents[i].name);
		}
		break;
	case 'a':
		if(argv[2][1] != '8' && argv[2][1] != '9') {
			fprintf(stderr, "Error: parameter N out of range\n");
			break;
		}
		printf("start extracting\n");
		for(i = 0; i < files; ++i) {
			printf("Extracting %s...", filents[i].name);
			if(th_extract_file(fp, &filents[i], argv[2][1] - '8'))
				printf("succeeded\n");
			else
				printf("failed\n");
			free(filents[i].name);
		}
		break;
	case 'e':
		if(argv[2][1] != '8' && argv[2][1] != '9') {
			fprintf(stderr, "Error: parameter N out of range\n");
			break;
		}
		if(argc < 4) {
			fprintf(stderr, "Error: not enough parameters\n");
			break;
		}
		for(i = 0; i < files; ++i) {
			if(!strcmp(filents[i].name, argv[3])) {
				printf("Extracting %s...", filents[i].name);
				if(th_extract_file(fp, &filents[i], argv[2][1] - '8'))
					printf("succeeded\n");
				else
					printf("failed\n");
			}
			free(filents[i].name);
		}
		break;
	}

	fclose(fp);
	return 0;
}
